@@ -1,12 +1,12 @@
 /* This is the machine-specific part for the 68000 and family
 
-  Copyright (C) 1995,1996,1997,1998,2000,2003 Free Software Foundation, Inc.
+  Copyright (C) 1995,1996,1997,1998,2000,2003,2005,2007,2014 Free Software Foundation, Inc.
 
   This file is part of Gforth.
 
   Gforth is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
-  as published by the Free Software Foundation; either version 2
+  as published by the Free Software Foundation, either version 3
   of the License, or (at your option) any later version.
 
   This program is distributed in the hope that it will be useful,
@@ -15,18 +15,9 @@
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
-  along with this program; if not, write to the Free Software
-  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
+  along with this program; if not, see http://www.gnu.org/licenses/.
 */
 
-#ifndef THREADING_SCHEME
-#define THREADING_SCHEME 3
-#endif
-
-/* define this if the processor cannot exploit instruction-level
-   parallelism (no pipelining or too few registers) */
-#define CISC_NEXT
-
 #if !defined USE_NO_TOS && !defined USE_TOS
 #define USE_TOS
 #endif
@@ -39,7 +30,16 @@
  */
 #if defined(apollo)
 #  define FLUSH_ICACHE(addr,size)    cache_$clear()
-#elif defined(NeXT)
+#elif defined(__NetBSD__)
+#  define FLUSH_ICACHE(addr,size)	do {				\
+		register void *addr_ asm("a1") = (addr);		\
+		register int size_ asm("d1") = (size);			\
+		register int cmd_ asm("d0") = 0x80000004;		\
+		asm volatile("	trap	#12"				\
+			: "=a" (addr_), "=d" (size_), "=d" (cmd_)	\
+			: "0" (addr_), "1" (size_), "2" (cmd_) : "a0");	\
+	} while (0)
+#elif defined(NeXT) || defined(sun)
 #  define FLUSH_ICACHE(addr,size)     asm("trap #2");
 #elif defined(hpux)
 #  include <sys/cache.h>
@@ -61,6 +61,12 @@ extern int cacheflush(void *, int, int, size_t);
 #  warning CODE words will not work.
 #endif
 
+#define ASM_SM_SLASH_REM(d1lo, d1hi, n1, n2, n3) \
+  asm("divs%.l %2,%1:%0": "=d"(n3),"=d"(n2) : "dmi"(n1),"0"(d1lo),"1"(d1hi):"cc")
+
+#define ASM_UM_SLASH_MOD(d1lo, d1hi, n1, n2, n3) \
+  asm("divu%.l %2,%1:%0": "=d"(n3),"=d"(n2) : "dmi"(n1),"0"(d1lo),"1"(d1hi):"cc")
+
 #ifdef FORCE_REG /* highly recommended */
 #if defined(amigaos)
 #  define IPREG asm("%a6")
