@@ -1,12 +1,12 @@
 /* Gforth virtual machine (aka inner interpreter)
 
-  Copyright (C) 1995,1996,1997,1998,2000,2003 Free Software Foundation, Inc.
+  Copyright (C) 1995,1996,1997,1998,2000,2003,2004,2005,2006,2007,2008,2010,2011,2012,2013,2014 Free Software Foundation, Inc.
 
   This file is part of Gforth.
 
   Gforth is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
-  as published by the Free Software Foundation; either version 2
+  as published by the Free Software Foundation, either version 3
   of the License, or (at your option) any later version.
 
   This program is distributed in the hope that it will be useful,
@@ -15,10 +15,15 @@
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
-  along with this program; if not, write to the Free Software
-  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
+  along with this program; if not, see http://www.gnu.org/licenses/.
 */
 
+#if defined(GFORTH_DEBUGGING) || defined(INDIRECT_THREADED) || defined(DOUBLY_INDIRECT) || defined(VM_PROFILING)
+#define USE_NO_TOS
+#else
+#define USE_TOS
+#endif
+
 #include "config.h"
 #include "forth.h"
 #include <ctype.h>
@@ -39,6 +44,9 @@
 #include <unistd.h>
 #include <pwd.h>
 #include <dirent.h>
+#ifdef HAVE_WCHAR_H
+#include <wchar.h>
+#endif
 #include <sys/resource.h>
 #ifdef HAVE_FNMATCH_H
 #include <fnmatch.h>
@@ -46,7 +54,7 @@
 #include "fnmatch.h"
 #endif
 #else
-#include "systypes.h"
+/* #include <systypes.h> */
 #endif
 
 #if defined(HAVE_LIBDL) || defined(HAVE_DLOPEN) /* what else? */
@@ -64,6 +72,11 @@
 #include <callback.h>
 #endif
 
+#ifdef HAS_DEBUG
+extern int debug;
+# define debugp(x...) do { if (debug) fprintf(x); } while (0)
+#endif
+
 #ifndef SEEK_SET
 /* should be defined in stdio.h, but some systems don't have it */
 #define SEEK_SET 0
@@ -81,11 +94,31 @@
 
 #ifdef MEMCMP_AS_SUBROUTINE
 extern int gforth_memcmp(const char * s1, const char * s2, size_t n);
+extern Char *gforth_memmove(Char * dest, const Char* src, Cell n);
+extern Char *gforth_memset(Char * s, Cell c, UCell n);
+extern Char *gforth_memcpy(Char * dest, const Char* src, Cell n);
 #define memcmp(s1,s2,n) gforth_memcmp(s1,s2,n)
+#define memmove(a,b,c) gforth_memmove(a,b,c)
+#define memset(a,b,c) gforth_memset(a,b,c)
+#define memcpy(a,b,c) gforth_memcpy(a,b,c)
 #endif
 
 #define NEWLINE	'\n'
 
+/* These two flags control whether divisions are checked by software.
+   The CHECK_DIVISION_SW is for those cases where the event is a
+   division by zero or overflow on the C level, and might be reported
+   by hardware; we might check forr that in autoconf and set the
+   switch appropriately, but currently don't.  The CHECK_DIVISION flag
+   is for the other cases. */
+#ifdef GFORTH_DEBUGGING
+#define CHECK_DIVISION_SW 1
+#define CHECK_DIVISION 1
+#else
+#define CHECK_DIVISION_SW 0
+#define CHECK_DIVISION 0
+#endif
+
 /* conversion on fetch */
 
 #define vm_Cell2f(_cell,_x)		((_x)=(Bool)(_cell))
@@ -144,6 +177,9 @@ extern int gforth_memcmp(const char * s1, const char * s2, size_t n);
 #ifndef LPREG
 #define LPREG
 #endif
+#ifndef CAREG
+#define CAREG
+#endif
 #ifndef CFAREG
 #define CFAREG
 #endif
@@ -153,9 +189,36 @@ extern int gforth_memcmp(const char * s1, const char * s2, size_t n);
 #ifndef TOSREG
 #define TOSREG
 #endif
+#ifndef spbREG
+#define spbREG
+#endif
+#ifndef spcREG
+#define spcREG
+#endif
+#ifndef spdREG
+#define spdREG
+#endif
+#ifndef speREG
+#define speREG
+#endif
+#ifndef spfREG
+#define spfREG
+#endif
+#ifndef spgREG
+#define spgREG
+#endif
+#ifndef sphREG
+#define sphREG
+#endif
 #ifndef FTOSREG
 #define FTOSREG
 #endif
+#ifndef OPREG
+#define OPREG
+#endif
+#ifndef SPSREG
+#define SPSREG
+#endif
 
 #ifndef CPU_DEP1
 # define CPU_DEP1 0
@@ -175,96 +238,161 @@ extern int gforth_memcmp(const char * s1, const char * s2, size_t n);
 #endif
 #define SUPER_CONTINUE
 
+#ifdef ASMCOMMENT
+/* an individualized asm statement so that (hopefully) gcc's optimizer
+   does not do cross-jumping */
+#define asmcomment(string) asm(ASMCOMMENT string)
+#else
+/* we don't know how to do an asm comment, so we just do an empty asm */
+#define asmcomment(string) asm("")
+#endif
+
+#define DEPTHOFF 0
 #ifdef GFORTH_DEBUGGING
+#if DEBUG
+#define NAME(string) if(debug) { saved_ip=ip; asmcomment(string); fprintf(stderr,"%08lx depth=%3ld tos=%016lx: "string"\n",(Cell)ip,((user_area*)up)->sp0+DEPTHOFF-sp,sp[0]);}
+#else /* !DEBUG */
 #define NAME(string) { saved_ip=ip; asm(""); }
 /* the asm here is to avoid reordering of following stuff above the
    assignment; this is an old-style asm (no operands), and therefore
    is treated like "asm volatile ..."; i.e., it prevents most
    reorderings across itself.  We want the assignment above first,
    because the stack loads may already cause a stack underflow. */
+#endif /* !DEBUG */
 #elif DEBUG
-#       define  NAME(string)    fprintf(stderr,"%08lx depth=%3ld: "string"\n",(Cell)ip,sp0+3-sp);
+#       define  NAME(string)    if(debug) {Cell __depth=((user_area*)up)->sp0+DEPTHOFF-sp; int i; fprintf(stderr,"%08lx depth=%3ld: "string,(Cell)ip,((user_area*)up)->sp0+DEPTHOFF-sp); for (i=__depth-1; i>0; i--) fprintf(stderr, " $%lx",sp[i]); fprintf(stderr, " $%lx\n",spTOS); }
+#elif ASMNAME
+#	define	NAME(string) asmcomment(string);
 #else
-#	define	NAME(string)
+#	define  NAME(string)
 #endif
 
 #ifdef DEBUG
 #define CFA_TO_NAME(__cfa) \
-      Cell len, i; \
-      char * name = __cfa; \
-      for(i=0; i<32; i+=sizeof(Cell)) { \
-        len = ((Cell*)name)[-1]; \
-        if(len < 0) { \
-	  len &= 0x1F; \
-          if((len+sizeof(Cell)) > i) break; \
-	} len = 0; \
-	name -= sizeof(Cell); \
-      }
+  Cell len = LONGNAME_COUNT(__cfa); \
+  char * name = LONGNAME_NAME(__cfa);
 #endif
 
-#ifdef HAS_FFCALL
-#define SAVE_REGS IF_spTOS(sp[0]=spTOS); IF_fpTOS(fp[0]=fpTOS); SP=sp; FP=fp; RP=rp; LP=lp;
-#define REST_REGS sp=SP; fp=FP; rp=RP; lp=LP; IF_spTOS(spTOS=sp[0]); IF_fpTOS(fpTOS=fp[0]);
+#ifdef STANDALONE
+jmp_buf * throw_jmp_handler;
+
+void throw(int code)
+{
+  longjmp(*throw_jmp_handler,code); /* !! or use siglongjmp ? */
+}
+#endif
+
+#if defined(HAS_FFCALL) || defined(HAS_LIBFFI)
+#define SAVE_REGS IF_fpTOS(fp[0]=fpTOS); gforth_SP=sp; gforth_FP=fp; gforth_RP=rp; gforth_LP=lp;
+#define REST_REGS sp=gforth_SP; fp=gforth_FP; rp=gforth_RP; lp=gforth_LP; IF_fpTOS(fpTOS=fp[0]);
 #endif
 
 #if !defined(ENGINE)
 /* normal engine */
 #define VARIANT(v)	(v)
 #define JUMP(target)	goto I_noop
-#define LABEL(name) J_##name: asm(""); I_##name:
+#define LABEL(name) H_##name: asm(ASMCOMMENT "I " #name); \
+     I_##name:
+#define LABEL_UU(name) H_##name: MAYBE_UNUSED asm(ASMCOMMENT "I " #name); \
+     I_##name: MAYBE_UNUSED
+#define LABEL3(name) J_##name: asm(ASMCOMMENT "J " #name);
+#define LABEL3_UU(name) J_##name: MAYBE_UNUSED asm(ASMCOMMENT "J " #name);
 
 #elif ENGINE==2
 /* variant with padding between VM instructions for finding out
    cross-inst jumps (for dynamic code) */
-#define engine engine2
+#define gforth_engine gforth_engine2
 #define VARIANT(v)	(v)
 #define JUMP(target)	goto I_noop
-#define LABEL(name) J_##name: SKIP16; I_##name:
-#define IN_ENGINE2
+#define LABEL(name) H_##name: asm(ASMCOMMENT "H " #name); SKIP16; \
+    asm(ASMCOMMENT "I " #name); I_##name:
+#define LABEL_UU(name) H_##name: MAYBE_UNUSED asm(ASMCOMMENT "H " #name); SKIP16; \
+    asm(ASMCOMMENT "I " #name); I_##name: MAYBE_UNUSED
+/* the SKIP16 after LABEL3 is there, because the ARM gcc may place
+   some constants after the final branch, and may refer to them from
+   the code before label3.  Since we don't copy the constants, we have
+   to make sure that such code is recognized as non-relocatable. */
+#define LABEL3(name) J_##name: asm(ASMCOMMENT "J " #name); SKIP16;
+#define LABEL3_UU(name) J_##name: MAYBE_UNUSED asm(ASMCOMMENT "J " #name); SKIP16;
 
 #elif ENGINE==3
 /* variant with different immediate arguments for finding out
    immediate arguments (for native code) */
-#define engine engine3
+#define gforth_engine gforth_engine3
 #define VARIANT(v)	((v)^0xffffffff)
 #define JUMP(target)	goto K_lit
-#define LABEL(name) J_##name: asm(""); I_##name:
+#define LABEL(name) H_##name: asm(ASMCOMMENT "I " #name); \
+    I_##name:
+#define LABEL_UU(name) H_##name: MAYBE_UNUSED asm(ASMCOMMENT "I " #name); \
+    I_##name: MAYBE_UNUSED 
+#define LABEL3(name) J_##name: asm(ASMCOMMENT "J " #name);
+#define LABEL3_UU(name) J_##name: MAYBE_UNUSED asm(ASMCOMMENT "J " #name);
 #else
 #error illegal ENGINE value
 #endif /* ENGINE */
 
-#define LABEL2(name) K_##name:
+/* the asm(""); is there to get a stop compiled on Itanium */
+#define LABEL2(name) K_##name: asm(ASMCOMMENT "K " #name);
+#define LABEL2_UU(name) K_##name: MAYBE_UNUSED asm(ASMCOMMENT "K " #name);
 
-Label *engine(Xt *ip0, Cell *sp0, Cell *rp0, Float *fp0, Address lp0)
+Label *gforth_engine(Xt *ip0 sr_proto)
 /* executes code at ip, if ip!=NULL
    returns array of machine code labels (for use in a loader), if ip==NULL
 */
 {
-#ifndef GFORTH_DEBUGGING
+  register stackpointers * SPs SPSREG = in_SPs;
+#undef gforth_SP
+#undef gforth_RP
+#undef gforth_LP
+#undef gforth_UP
+#define gforth_SP (SPs->spx)
+#define gforth_RP (SPs->rpx)
+#define gforth_LP (SPs->lpx)
+#define gforth_UP (SPs->upx)
+#if defined(GFORTH_DEBUGGING)
+# undef saved_ip
+# define rp (SPs->s_rp)
+# define saved_ip (SPs->s_ip)
+#else /* !defined(GFORTH_DEBUGGING) */
   register Cell *rp RPREG;
-#endif
+# undef saved_ip
+  Xt* MAYBE_UNUSED saved_ip;
+#endif /* !defined(GFORTH_DEBUGGING) */
 #ifndef NO_IP
   register Xt *ip IPREG = ip0;
 #endif
-  register Cell *sp SPREG = sp0;
-  register Float *fp FPREG = fp0;
-  register Address lp LPREG = lp0;
+  register Cell *sp SPREG = gforth_SP;
+  register Float *fp FPREG = gforth_FP;
+  register Address lp LPREG = gforth_LP;
   register Xt cfa CFAREG;
+  register Label real_ca CAREG;
+#ifdef HAS_OBJECTS
+  register Char * op OPREG = NULL;
+#endif
 #ifdef MORE_VARS
   MORE_VARS
 #endif
 #ifdef HAS_FFCALL
   av_alist alist;
-  extern va_alist clist;
+  extern va_alist gforth_clist;
   float frv;
   int irv;
   double drv;
   long long llrv;
   void * prv;
 #endif
-  register Address up UPREG = UP;
-  IF_spTOS(register Cell spTOS TOSREG;)
+  register user_area* up UPREG = gforth_UP;
+#if !defined(GFORTH_DEBUGGING)
+  register Cell MAYBE_UNUSED spTOS TOSREG;
+  register Cell MAYBE_UNUSED spb spbREG;
+  register Cell MAYBE_UNUSED spc spcREG;
+  register Cell MAYBE_UNUSED spd spdREG;
+  register Cell MAYBE_UNUSED spe speREG;
+  register Cell MAYBE_UNUSED spf spfREG;
+  register Cell MAYBE_UNUSED spg spgREG;
+  register Cell MAYBE_UNUSED sph sphREG;
   IF_fpTOS(register Float fpTOS FTOSREG;)
+#endif /* !defined(GFORTH_DEBUGGING) */
 #if defined(DOUBLY_INDIRECT)
   static Label *symbols;
   static void *routines[]= {
@@ -274,52 +402,88 @@ Label *engine(Xt *ip0, Cell *sp0, Cell *rp0, Float *fp0, Address lp0)
 #define MAX_SYMBOLS (sizeof(symbols)/sizeof(symbols[0]))
 #endif /* !defined(DOUBLY_INDIRECT) */
 #define INST_ADDR(name) ((Label)&&I_##name)
-#include "prim_lab.i"
+#include PRIM_LAB_I
 #undef INST_ADDR
     (Label)0,
 #define INST_ADDR(name) ((Label)&&K_##name)
-#include "prim_lab.i"
+#include PRIM_LAB_I
 #undef INST_ADDR
 #define INST_ADDR(name) ((Label)&&J_##name)
-#include "prim_lab.i"
+#include PRIM_LAB_I
+#undef INST_ADDR
+    (Label)&&after_last,
+    (Label)&&before_goto,
+    (Label)&&after_goto,
+/* just mention the H_ labels, so the SKIP16s are not optimized away */
+#define INST_ADDR(name) ((Label)&&H_##name)
+#include PRIM_LAB_I
 #undef INST_ADDR
-    (Label)&&after_last
   };
+#ifdef STANDALONE
+#define INST_ADDR(name) ((Label)&&I_##name)
+#include "image.i"
+#undef INST_ADDR
+#endif
 #ifdef CPU_DEP2
   CPU_DEP2
 #endif
 
-  rp = rp0;
+  rp = SPs->rpx;
 #ifdef DEBUG
-  fprintf(stderr,"ip=%x, sp=%x, rp=%x, fp=%x, lp=%x, up=%x\n",
-          (unsigned)ip0,(unsigned)sp,(unsigned)rp,
-	  (unsigned)fp,(unsigned)lp,(unsigned)up);
+  debugp(stderr,"ip=%lx, sp=%lx, rp=%lx, fp=%lx, lp=%lx, up=%lx\n",
+	 (Cell)ip0,(Cell)sp,(Cell)rp,
+	 (Cell)fp,(Cell)lp,(Cell)up);
 #endif
 
   if (ip0 == NULL) {
 #if defined(DOUBLY_INDIRECT)
 #define CODE_OFFSET (26*sizeof(Cell))
 #define XT_OFFSET (22*sizeof(Cell))
+#define LABEL_OFFSET (18*sizeof(Cell))
     int i;
     Cell code_offset = offset_image? CODE_OFFSET : 0;
     Cell xt_offset = offset_image? XT_OFFSET : 0;
+    Cell label_offset = offset_image? LABEL_OFFSET : 0;
+
+    debugp(stderr, "offsets code/xt/label: %lx/%lx/%lx\n",
+	   code_offset, xt_offset, label_offset);
 
     symbols = (Label *)(malloc(MAX_SYMBOLS*sizeof(Cell)+CODE_OFFSET)+code_offset);
     xts = (Label *)(malloc(MAX_SYMBOLS*sizeof(Cell)+XT_OFFSET)+xt_offset);
-    for (i=0; i<DOESJUMP+1; i++)
+    labels = (Label *)(malloc(MAX_SYMBOLS*sizeof(Cell)+LABEL_OFFSET)+label_offset);
+    
+    for (i=0; i<DOER_MAX+1; i++) {
+      labels[i] = routines[i];
       xts[i] = symbols[i] = (Label)routines[i];
+    }
     for (; routines[i]!=0; i++) {
       if (i>=MAX_SYMBOLS) {
 	fprintf(stderr,"gforth-ditc: more than %ld primitives\n",(long)MAX_SYMBOLS);
 	exit(1);
       }
-      xts[i] = symbols[i] = &routines[i];
+      labels[i] = routines[i];
+      xts[i] = symbols[i] = (Label)&labels[i];
     }
 #endif /* defined(DOUBLY_INDIRECT) */
+#ifdef STANDALONE
+    return image;
+#else
     return symbols;
+#endif
   }
 
-  IF_spTOS(spTOS = sp[0]);
+#ifdef USE_TOS
+  sp += STACK_CACHE_DEFAULT-1;
+  /* some of those registers are dead, but its simpler to initialize them all */  spTOS = sp[0];
+  spb = sp[-1];
+  spc = sp[-2];
+  spd = sp[-3];
+  spe = sp[-4];
+  spf = sp[-5];
+  spg = sp[-6];
+  sph = sp[-7];
+#endif
+
   IF_fpTOS(fpTOS = fp[0]);
 /*  prep_terminal(); */
 #ifdef NO_IP
@@ -334,7 +498,11 @@ Label *engine(Xt *ip0, Cell *sp0, Cell *rp0, Float *fp0, Address lp0)
   CPU_DEP3
 #endif
 
-#include "prim.i"
-  after_last: return (Label *)0;
+#include PRIM_I
+  after_last: 
   /*needed only to get the length of the last primitive */
+  FIRST_NEXT;
+  LABEL_UU(return_0)
+  NEXT; /* gcc-4.9 workaround: Requires yet another dummy NEXT */
+  return 0;
 }
