@@ -1,12 +1,12 @@
 /* a strerror implemenation
 
-  Copyright (C) 1995,2000,2003 Free Software Foundation, Inc.
+  Copyright (C) 1995,2000,2003,2007,2014 Free Software Foundation, Inc.
 
   This file is part of Gforth.
 
   Gforth is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
-  as published by the Free Software Foundation; either version 2
+  as published by the Free Software Foundation, either version 3
   of the License, or (at your option) any later version.
 
   This program is distributed in the hope that it will be useful,
@@ -15,15 +15,14 @@
   GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License
-  along with this program; if not, write to the Free Software
-  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
+  along with this program; if not, see http://www.gnu.org/licenses/.
 */
 
 #include <errno.h>
 #include <stdio.h>
 
-extern char *sys_errlist[];
-extern int sys_nerr;
+extern const char *const sys_errlist[];
+extern const int sys_nerr;
 
 char *strerror(int err)
 {
