@@ -1,6 +1,6 @@
 ;;; gforth.el --- major mode for editing (G)Forth sources
 
-;; Copyright (C) 1995,1996,1997,1998,2000,2001,2003 Free Software Foundation, Inc.
+;; Copyright (C) 1995,1996,1997,1998,2000,2001,2003,2004,2007,2008,2010,2011,2012,2013,2014 Free Software Foundation, Inc.
 
 ;; This file is part of Gforth.
 
@@ -59,15 +59,15 @@
 (if (not (boundp 'emacs-major-version))
     (defconst emacs-major-version
       (progn (string-match "^[0-9]+" emacs-version)
-	     (string-to-int (match-string 0 emacs-version)))))
-
-(defun forth-emacs-older (major minor)
-  (or (< emacs-major-version major)
-      (and (= emacs-major-version major) (< emacs-minor-version minor))))
+	     (string-to-number (match-string 0 emacs-version)))))
 
 ;; Code ripped from `subr.el' for compatability with Emacs versions
 ;; prior to 20.1
 (eval-when-compile 
+(defun forth-emacs-older (major minor)
+  (or (< emacs-major-version major)
+      (and (= emacs-major-version major) (< emacs-minor-version minor))))
+
   (if (forth-emacs-older 20 1)
       (progn
 	(defmacro when (cond &rest body)
@@ -79,8 +79,9 @@
 
 ;; `no-error' argument of require not supported in Emacs versions
 ;; prior to 20.4 :-(
+(eval-and-compile
 (defun forth-require (feature)
-  (condition-case err (require feature) (error nil)))
+    (condition-case err (require feature) (error nil))))
 
 (require 'font-lock)
 
@@ -236,27 +237,35 @@ PARSED-TYPE specifies what kind of text is parsed. It should be on of 'name',
 	(("]" "]l") definition-starter (font-lock-keyword-face . 1))
 	((":") definition-starter (font-lock-keyword-face . 1)
 	 "[ \t\n]" t name (font-lock-function-name-face . 3))
+	(("event:") definition-starter (font-lock-keyword-face . 1)
+	 "[ \t\n]" t name (font-lock-function-name-face . 3))
 	(("immediate" "compile-only" "restrict")
 	 immediate (font-lock-keyword-face . 1))
-	(("does>") compile-only (font-lock-keyword-face . 1))
-	((":noname") definition-starter (font-lock-keyword-face . 1))
-	((";" ";code") definition-ender (font-lock-keyword-face . 1))
+	(("does>") definition-starter (font-lock-keyword-face . 1))
+	((":noname" "comp:" "post:") definition-starter (font-lock-keyword-face . 1))
+	((";" ";code" ";abi-code") definition-ender (font-lock-keyword-face . 1))
 	(("include" "require" "needs" "use") 
 	 non-immediate (font-lock-keyword-face . 1) 
 	 "[\n\t ]" t string (font-lock-string-face . 1))
 	(("included" "required" "thru" "load")
 	 non-immediate (font-lock-keyword-face . 1))
+	(("code" "abi-code")
+	 non-immediate (font-lock-keyword-face . 1)
+	 "[ \t\n]" t name (font-lock-function-name-face . 3))
+	(("end-code")
+	 non-immediate (font-lock-keyword-face . 1))
 	(("[char]") compile-only (font-lock-keyword-face . 1)
 	 "[ \t\n]" t string (font-lock-string-face . 1))
 	(("char") non-immediate (font-lock-keyword-face . 1)
 	 "[ \t\n]" t string (font-lock-string-face . 1))
-	(("s\"" "c\"") immediate (font-lock-string-face . 1)
+	("'.'?" non-immediate (font-lock-string-face . 1))
+	(("s\"" "c\"" "s\\\"") immediate (font-lock-string-face . 1)
 	 "[\"\n]" nil string (font-lock-string-face . 1))
-	((".\"") compile-only (font-lock-string-face . 1)
+	((".\"" ".\\\"") compile-only (font-lock-string-face . 1)
 	 "[\"\n]" nil string (font-lock-string-face . 1))
 	(("abort\"") compile-only (font-lock-keyword-face . 1)
 	 "[\"\n]" nil string (font-lock-string-face . 1))
-	(("{") compile-only (font-lock-variable-name-face . 1)
+	(("{" "{:") compile-only (font-lock-variable-name-face . 1)
 	 "[\n}]" nil name (font-lock-variable-name-face . 1))
 	((".(" "(") immediate (font-lock-comment-face . 1)
 	  ")" nil comment (font-lock-comment-face . 1))
@@ -265,22 +274,23 @@ PARSED-TYPE specifies what kind of text is parsed. It should be on of 'name',
 	  
 	(("[if]" "[?do]" "[do]" "[for]" "[begin]" 
 	  "[endif]" "[then]" "[loop]" "[+loop]" "[next]" "[until]" "[repeat]"
-	  "[again]" "[while]" "[else]")
+	  "[again]" "[while]" "[else]" "[:" ";]" "nope")
 	 immediate (font-lock-keyword-face . 2))
-	(("[ifdef]" "[ifundef]") immediate (font-lock-keyword-face . 2)
+	(("[ifdef]" "[ifundef]" "[defined]" "[undefined]") immediate (font-lock-keyword-face . 2)
 	 "[ \t\n]" t name (font-lock-function-name-face . 3))
 	(("if" "begin" "ahead" "do" "?do" "+do" "u+do" "-do" "u-do" "for" 
 	  "case" "of" "?dup-if" "?dup-0=-if" "then" "endif" "until"
 	  "repeat" "again" "leave" "?leave"
 	  "loop" "+loop" "-loop" "next" "endcase" "endof" "else" "while" "try"
-	  "recover" "endtry" "assert(" "assert0(" "assert1(" "assert2(" 
+	  "recover" "endtry" "iferror" "restore" "endtry-iferror"
+	  "assert(" "assert0(" "assert1(" "assert2("
 	  "assert3(" ")" "<interpretation" "<compilation" "interpretation>" 
 	  "compilation>")
 	 compile-only (font-lock-keyword-face . 2))
 
 	(("true" "false" "c/l" "bl" "cell" "pi" "w/o" "r/o" "r/w") 
 	 non-immediate (font-lock-constant-face . 2))
-	(("~~" "break:" "dbg") compile-only (font-lock-warning-face . 2))
+	(("~~" "break:" "dbg" "???" "wtf??" "!!fixme!!" "~~bt" "~~1bt" "bt") compile-only (font-lock-warning-face . 2))
 	(("break\"") compile-only (font-lock-warning-face . 1)
 	 "[\"\n]" nil string (font-lock-string-face . 1))
 	(("postpone" "[is]" "defers" "[']" "[compile]") 
@@ -298,8 +308,10 @@ PARSED-TYPE specifies what kind of text is parsed. It should be on of 'name',
 	 "[ \t\n]" t name (font-lock-variable-name-face . 3))
 
 	(("create" "variable" "constant" "2variable" "2constant" "fvariable"
-	  "fconstant" "value" "field" "user" "vocabulary" 
-	  "create-interpret/compile")
+	  "fconstant" "value" "2value" "field" "user" "vocabulary" 
+	  "create-interpret/compile" "interpret/compile:"
+	  "debug:" "field:" "2field:" "ffield:" "sffield:" "dffield:"
+	  "uvar" "uvalue" "cfield:" "wfield:" "lfield:")
 	 non-immediate (font-lock-type-face . 2)
 	 "[ \t\n]" t name (font-lock-variable-name-face . 3))
 	("\\S-+%" non-immediate (font-lock-type-face . 2))
@@ -308,9 +320,32 @@ PARSED-TYPE specifies what kind of text is parsed. It should be on of 'name',
 	 "[ \t\n]" t name (font-lock-function-name-face . 3))
 	(("end-struct") non-immediate (font-lock-keyword-face . 2)
 	 "[ \t\n]" t name (font-lock-type-face . 3))
-	(("struct") non-immediate (font-lock-keyword-face . 2))
+	(("struct" "end-c-library" "c-library-name" "end-structure") 
+	 non-immediate (font-lock-keyword-face . 2))
+	(("c-library" "begin-structure") non-immediate (font-lock-keyword-face . 2)
+	 "[ \t\n]" t name (font-lock-variable-name-face . 3))
+	(("c-variable") non-immediate (font-lock-type-face . 1)
+	 "[ \t\n]" t name (font-lock-function-name-face . 3)
+	 "[ \t\n]" t name (font-lock-function-name-face . 3))
+	(("c-function" "c-value") non-immediate (font-lock-type-face . 1)
+	 "[ \t\n]" t name (font-lock-function-name-face . 3)
+	 "[ \t\n]" t name (font-lock-function-name-face . 3)
+	 "[\n]" nil comment (font-lock-variable-name-face . 3))
+	(("c-callback") non-immediate (font-lock-type-face . 1)
+	 "[ \t\n]" t name (font-lock-function-name-face . 3)
+	 "[\n]" nil comment (font-lock-variable-name-face . 3))
+	(("c-funptr") non-immediate (font-lock-type-face . 1)
+	 "[ \t\n]" t name (font-lock-function-name-face . 3)
+	 "[}]" t name (font-lock-function-name-face . 3)
+	 "[\n]" nil comment (font-lock-variable-name-face . 3))
+	(("\\c") non-immediate (font-lock-keyword-face . 1)
+	 "[\n]" nil string (font-lock-string-face . 1))
 	("-?[0-9]+\\(\\.[0-9]*e\\(-?[0-9]+\\)?\\|\\.?[0-9a-f]*\\)" 
 	 immediate (font-lock-constant-face . 3))
+	("-?\\([&#][0-9]+\\|\\(0x\\|\\$\\)[0-9a-f.]+\\|%[01]+\\)"
+	 immediate (font-lock-constant-face . 3))
+	("[a-z\-0-9]+(" immediate (font-lock-comment-face . 1)
+	 ")" nil comment (font-lock-comment-face . 1))
 	))
 
 (defvar forth-use-objects nil 
@@ -415,27 +450,28 @@ INDENT1 and INDENT2 are indentation specifications of the form
 
 (setq forth-indent-words
       '((("if" "begin" "do" "?do" "+do" "-do" "u+do"
-	  "u-do" "?dup-if" "?dup-0=-if" "case" "of" "try" 
-	  "[if]" "[ifdef]" "[ifundef]" "[begin]" "[for]" "[do]" "[?do]")
+	  "u-do" "?dup-if" "?dup-0=-if" "case" "of" "try" "iferror"
+	  "[if]" "[ifdef]" "[ifundef]" "[begin]" "[for]" "[do]" "[?do]" "[:")
 	 (0 . 2) (0 . 2))
-	((":" ":noname" "code" "struct" "m:" ":m" "class" "interface")
+	((":" ":noname" "code" "abi-code" "struct" "m:" ":m" "class" 
+	  "interface" "c-library" "c-library-name" "comp:" "post:" "begin-structure")
 	 (0 . 2) (0 . 2) non-immediate)
 	("\\S-+%$" (0 . 2) (0 . 0) non-immediate)
 	((";" ";m") (-2 . 0) (0 . -2))
 	(("again" "then" "endif" "endtry" "endcase" "endof" 
 	  "[then]" "[endif]" "[loop]" "[+loop]" "[next]" 
-	  "[until]" "[again]" "loop")
+	  "[until]" "[again]" "loop" ";]" "nope")
 	 (-2 . 0) (0 . -2))
 	(("end-code" "end-class" "end-interface" "end-class-noname" 
-	  "end-interface-noname" "end-struct" "class;")
+	  "end-interface-noname" "end-struct" "class;" "end-c-library" "end-structure")
 	 (-2 . 0) (0 . -2) non-immediate)
 	(("protected" "public" "how:") (-1 . 1) (0 . 0) non-immediate)
 	(("+loop" "-loop" "until") (-2 . 0) (-2 . 0))
-	(("else" "recover" "[else]") (-2 . 2) (0 . 0))
-	(("does>") (-1 . 1) (0 . 0))
+	(("else" "recover" "restore" "endtry-iferror" "[else]")
+	 (-2 . 2) (0 . 0))
+	(("does>" "compile>" "int>" ";code" ";abi-code") (-1 . 1) (0 . 0))
 	(("while" "[while]") (-2 . 4) (0 . 2))
-	(("repeat" "[repeat]") (-4 . 0) (0 . -4))
-	(("\\g") (-2 . 2) (0 . 0))))
+	(("repeat" "[repeat]") (-4 . 0) (0 . -4))))
 
 (defvar forth-local-indent-words nil 
   "List of Forth words to prepend to `forth-indent-words', when a forth-mode
@@ -460,12 +496,11 @@ End:\" construct).")
 ;; in Lisp??
 (defun forth-filter (predicate list)
   (let ((filtered nil))
-    (mapcar (lambda (item)
+    (dolist (item list)
 	      (when (funcall predicate item)
 		(if filtered
 		    (nconc filtered (list item))
-		  (setq filtered (cons item nil))))
-	      nil) list)
+          (setq filtered (cons item nil)))))
     filtered))
 
 ;; Helper function for `forth-compile-word': return whether word has to be
@@ -486,7 +521,7 @@ End:\" construct).")
 	 (regexp 
 	  (concat "\\(" (cond ((stringp matcher) matcher)
 			      ((listp matcher) (regexp-opt matcher))
-			      (t (error "Invalid matcher `%s'")))
+			      (t (error "Invalid matcher")))
 		  "\\)"))
 	 (depth (regexp-opt-depth regexp))
 	 (description (cdr word)))
@@ -509,9 +544,7 @@ End:\" construct).")
 			sub 
 			)) 
 		    mapped)))
-    (let ((result (cons regexp sub-list)))
-      (byte-compile 'result)
-      result)))
+    (cons regexp sub-list)))
 
 (defun forth-compile-words ()
   "Compile the the words from `forth-words' and `forth-indent-words' into
@@ -528,6 +561,8 @@ End:\" construct).")
   "Parse and bind local variables, set in the contents of the current 
  forth-mode buffer. Prepend `forth-local-words' to `forth-words' and 
  `forth-local-indent-words' to `forth-indent-words'."
+  (put 'forth-local-indent-words 'safe-local-variable 'listp)
+  (put 'forth-local-words 'safe-local-variable 'listp)
   (hack-local-variables)
   (setq forth-words (append forth-local-words forth-words))
   (setq forth-indent-words (append forth-local-indent-words 
@@ -697,14 +732,14 @@ End:\" construct).")
 (eval-when-compile 
   (defmacro forth-save-buffer-state (varlist &rest body)
     "Bind variables according to VARLIST and eval BODY restoring buffer state."
-    (` (let* ((,@ (append varlist
+    `(let* (,@(append varlist
 		   '((modified (buffer-modified-p)) (buffer-undo-list t)
 		     (inhibit-read-only t) (inhibit-point-motion-hooks t)
 		     before-change-functions after-change-functions
-		     deactivate-mark buffer-file-name buffer-file-truename))))
-	 (,@ body)
+                        deactivate-mark buffer-file-name buffer-file-truename)))
+       ,@body
 	 (when (and (not modified) (buffer-modified-p))
-	   (set-buffer-modified-p nil))))))
+         (set-buffer-modified-p nil)))))
 
 ;; Function that is added to the `change-functions' hook. Calls 
 ;; `forth-update-properties' and keeps care of disabling undo information
@@ -725,19 +760,19 @@ End:\" construct).")
 		 (get-text-property from 'fontified))
        (forth-update-properties from to)))))
 
-(eval-when-compile
-  (byte-compile 'forth-set-word-properties)
-  (byte-compile 'forth-next-known-forth-word)
-  (byte-compile 'forth-update-properties)
-  (byte-compile 'forth-delete-properties)
-  (byte-compile 'forth-get-regexp-branch)) 
+(byte-compile 'forth-set-word-properties)
+(byte-compile 'forth-next-known-forth-word)
+(byte-compile 'forth-update-properties)
+(byte-compile 'forth-delete-properties)
+(byte-compile 'forth-get-regexp-branch)
 
 ;;; imenu support
 ;;;
 (defvar forth-defining-words 
   '("VARIABLE" "CONSTANT" "2VARIABLE" "2CONSTANT" "FVARIABLE" "FCONSTANT"
-   "USER" "VALUE" "field" "end-struct" "VOCABULARY" "CREATE" ":" "CODE"
-   "DEFER" "ALIAS")
+   "USER" "VALUE" "2Value" "field" "end-struct" "VOCABULARY" "CREATE" ":" "CODE"
+   "DEFER" "ALIAS" "interpret/compile:" "debug:" "field:" "2field:" "ffield:"
+   "sffield:" "dffield:" "uvar" "uvalue")
   "List of words, that define the following word.
 Used for imenu index generation.")
 
@@ -827,8 +862,8 @@ Used for imenu index generation.")
 
 ;; Return the column increment, that the current line of forth code does to
 ;; the current or following lines. `which' specifies which indentation values
-;; to use. 0 means the indentation of following lines relative to current 
-;; line, 1 means the indentation of the current line relative to the previous 
+;; to use. 1 means the indentation of following lines relative to current 
+;; line, 0 means the indentation of the current line relative to the previous 
 ;; line. Return `nil', if there are no indentation words on the current line.
 (defun forth-get-column-incr (which)
   (save-excursion
@@ -1092,7 +1127,6 @@ exceeds 64 characters."
 (define-key forth-mode-map "\C-x\\" 'comment-region)
 (define-key forth-mode-map "\C-x~" 'forth-remove-tracers)
 (define-key forth-mode-map "\C-x\C-m" 'forth-split)
-(define-key forth-mode-map "\e " 'forth-reload)
 (define-key forth-mode-map "\t" 'forth-indent-command)
 (define-key forth-mode-map "\C-m" 'forth-reindent-then-newline-and-indent)
 (define-key forth-mode-map "\M-q" 'forth-fill-paragraph)
@@ -1365,7 +1399,7 @@ programmers who tend to fill code won't use emacs anyway:-)."
 (defun forth-remove-tracers ()
   "Remove tracers of the form `~~ '. Queries the user for each occurrence."
   (interactive)
-  (query-replace-regexp "\\(~~ \\| ~~$\\)" "" nil))
+  (query-replace-regexp "\\(~~[ \t]\\|[ \t]~~$\\)" "" nil))
 
 (define-key forth-mode-map "\C-x\C-e" 'compile)
 (define-key forth-mode-map "\C-x\C-n" 'next-error)
@@ -1383,18 +1417,12 @@ programmers who tend to fill code won't use emacs anyway:-)."
       (progn
 	(delete-other-windows)
 	(split-window-vertically
-	 (/ (* (screen-height) forth-percent-height) 100))
+	 (/ (frame-height) 2))
 	(other-window 1)
 	(switch-to-buffer buffer)
 	(goto-char (point-max))
 	(other-window 1))))
 
-(defun forth-compile (command)
-  (interactive (list (setq forth-compile-command (read-string "Compile command: " forth-compile-command))))
-  (forth-split-1 "*compilation*")
-  (setq ctools-compile-command command)
-  (compile1 ctools-compile-command "No more errors"))
-
 ;;; Forth menu
 ;;; Mikael Karlsson <qramika@eras70.ericsson.se>
 
@@ -1614,8 +1642,7 @@ of `forth-program-name').  Runs the hooks `inferior-forth-mode-hook'
 	(forth-end-of-paragraph)
 	(skip-chars-backward  "\t\n ")
 	(setq end (point))
-	(if (re-search-backward "\n[ \t]*\n" nil t)
-	    (setq start (point))
+	(if (null (re-search-backward "\n[ \t]*\n" nil t))
 	  (goto-char (point-min)))
 	(skip-chars-forward  "\t\n ")
 	(forth-send-region (point) end))))
@@ -1653,11 +1680,11 @@ With argument, position cursor at end of buffer."
 	   (push-mark)
 	   (goto-char (point-max)))))
 
-  (defun forth-send-region-and-go (start end)
+  (defun forth-send-region-and-go (my-start end)
     "Send the current region to the inferior Forth process.
 Then switch to the process buffer."
     (interactive "r")
-    (forth-send-region start end)
+    (forth-send-region my-start end)
     (forth-switch-to-interactive t))
 
   (defcustom forth-source-modes '(forth-mode forth-block-mode)
@@ -1680,12 +1707,10 @@ next one.")
 				    forth-source-modes t)) ; T because LOAD
 					; needs an exact name
     (comint-check-source file-name) ; Check to see if buffer needs saved.
-    (setq forth-prev-l/c-dir/file (cons (file-name-directory    file-name)
+    (setq forth-prev-l/c-dir/file (cons (file-name-directory file-name)
 					(file-name-nondirectory file-name)))
-    (comint-send-string (forth-proc) (concat "(load \""
-					     file-name
-					     "\"\)\n")))
-
+    (comint-send-string (forth-proc)
+			(concat "s\" " file-name "\" included\n")))
 
   
   (defvar forth-process-buffer nil "*The current Forth process buffer.
